use super::super::{Input, Word};
use super::{PatternParseError, PatternToken};
use std::fmt::{Debug, Display, Formatter};

enum BracketType {
    Optional,
    Choice,
}

#[derive(Clone)]
pub enum Pattern {
    Word(Word),
    Choice(Vec<Pattern>),
    Sequence(Vec<Pattern>),
    Optional(Box<Pattern>),
}
impl Pattern {
    pub fn matches(&self, input: &Input) -> bool {
        match self.consume_words(input.words()) {
            Ok(words) => words.is_empty(),
            Err(_) => false,
        }
    }

    fn consume_words<'a>(&self, words: &'a [Word]) -> Result<&'a [Word], ()> {
        match self {
            Pattern::Sequence(patterns) => {
                let mut words = words.clone();
                for pattern in patterns {
                    words = pattern.consume_words(words)?;
                }
                return Ok(words);
            }
            Pattern::Choice(patterns) => {
                for pattern in patterns {
                    if let Ok(words) = pattern.consume_words(words) {
                        return Ok(words);
                    }
                }
                return Err(());
            }
            Pattern::Optional(pattern) => match pattern.consume_words(words) {
                Ok(words) => return Ok(words),
                Err(_) => return Ok(words),
            },
            Pattern::Word(pattern_word) => {
                let first_word = match words.get(0) {
                    Some(word) => word,
                    None => return Err(()),
                };
                if pattern_word == first_word {
                    return Ok(&words[1..]);
                } else {
                    return Err(());
                }
            }
        }
    }

    pub fn parse(pattern_string: &str) -> Result<Self, PatternParseError> {
        let tokens = PatternToken::tokenise(pattern_string);
        Self::parse_tokens(&tokens)
    }

    fn parse_tokens(tokens: &[PatternToken]) -> Result<Self, PatternParseError> {
        let mut choice = Vec::new();
        let mut sequence = Vec::new();

        let mut bracket_stack = Vec::new();
        let mut outer_bracket_type = None;
        let mut bracket_start_index = 0;

        for (index, token) in tokens.iter().enumerate() {
            if bracket_stack.is_empty() {
                bracket_start_index = index;
            }
            match token {
                PatternToken::ChoiceStart => bracket_stack.push(BracketType::Choice),
                PatternToken::ChoiceEnd => match bracket_stack.pop() {
                    Some(BracketType::Choice) => outer_bracket_type = Some(BracketType::Choice),
                    Some(_) => return Err(PatternParseError::UnmatchedChoiceStart),
                    None => return Err(PatternParseError::UnmatchedChoiceEnd),
                },
                PatternToken::OptionalStart => bracket_stack.push(BracketType::Optional),
                PatternToken::OptionalEnd => match bracket_stack.pop() {
                    Some(BracketType::Optional) => outer_bracket_type = Some(BracketType::Optional),
                    Some(_) => return Err(PatternParseError::UnmatchedOptionalEnd),
                    None => return Err(PatternParseError::UnmatchedOptionalEnd),
                },
                _ => (),
            }
            if bracket_stack.is_empty() {
                if let Some(ref bracket_type) = outer_bracket_type {
                    let inner_tokens = &tokens[bracket_start_index + 1..index];
                    if inner_tokens.len() > 0 {
                        let pattern = Self::parse_tokens(inner_tokens)?;
                        sequence.push(match bracket_type {
                            BracketType::Optional => Pattern::Optional(Box::new(pattern)),
                            BracketType::Choice => pattern,
                        });
                    } else {
                        return Err(match bracket_type {
                            BracketType::Optional => PatternParseError::EmptyOptional,
                            BracketType::Choice => PatternParseError::EmptyChoice,
                        });
                    }
                    outer_bracket_type = None;
                } else {
                    match token {
                        PatternToken::Word(word) => sequence.push(Pattern::Word(word.clone())),
                        PatternToken::ChoiceSeparator => {
                            if !sequence.is_empty() {
                                let pattern = Pattern::Sequence(sequence.clone()).clean()?;
                                if let Pattern::Optional(_) = pattern {
                                    return Err(PatternParseError::OptionalInChoice);
                                };
                                choice.push(pattern);
                                sequence.clear();
                            }
                        }
                        _ => (),
                    }
                }
            }
        }
        if !bracket_stack.is_empty() {
            return Err(match bracket_stack[0] {
                BracketType::Optional => PatternParseError::UnmatchedOptionalStart,
                BracketType::Choice => PatternParseError::UnmatchedChoiceStart,
            });
        }
        if !choice.is_empty() && !sequence.is_empty() {
            // Flush the final choice
            let pattern = Pattern::Sequence(sequence.clone()).clean()?;
            if let Pattern::Optional(_) = pattern {
                return Err(PatternParseError::OptionalInChoice);
            };
            choice.push(pattern);
        }
        if !choice.is_empty() {
            Ok(Pattern::Choice(choice).clean()?)
        } else if !sequence.is_empty() {
            Ok(Pattern::Sequence(sequence).clean()?)
        } else {
            Err(PatternParseError::EmptySequence)
        }
    }

    fn clean(self) -> Result<Self, PatternParseError> {
        match self {
            Pattern::Choice(ref patterns) => {
                if patterns.len() == 0 {
                    Err(PatternParseError::EmptyChoice)
                } else if patterns.len() == 1 {
                    Ok(patterns[0].clone())
                } else {
                    Ok(self)
                }
            }
            Pattern::Sequence(ref patterns) => {
                if patterns.len() == 0 {
                    Err(PatternParseError::EmptySequence)
                } else if patterns.len() == 1 {
                    Ok(patterns[0].clone())
                } else {
                    Ok(self)
                }
            }
            _ => Ok(self),
        }
    }

    fn to_display_string(&self) -> String {
        let patterns_to_string = |v: &[Pattern], sep: &str| {
            v.iter()
                .map(|p| p.to_display_string())
                .collect::<Vec<String>>()
                .join(sep)
        };

        match self {
            Pattern::Choice(patterns) => {
                format!("({})", patterns_to_string(patterns, "|"))
            }
            Pattern::Sequence(patterns) => {
                format!("{}", patterns_to_string(patterns, " "))
            }
            Pattern::Optional(pattern) => format!("[{}]", pattern.to_display_string()),
            Pattern::Word(word) => format!("{}", word),
        }
    }
    fn to_debug_string(&self) -> String {
        let patterns_to_string = |v: &[Pattern]| {
            v.iter()
                .map(|p| p.to_debug_string())
                .collect::<Vec<String>>()
                .join(", ")
        };

        match self {
            Pattern::Choice(patterns) => {
                format!("Choice({})", patterns_to_string(patterns))
            }
            Pattern::Sequence(patterns) => {
                format!("Sequence({})", patterns_to_string(patterns))
            }
            Pattern::Optional(pattern) => format!("Optional({})", pattern.to_debug_string()),
            Pattern::Word(word) => format!("Word({})", word),
        }
    }
}

impl Display for Pattern {
    fn fmt(&self, f: &mut Formatter) -> Result<(), std::fmt::Error> {
        write!(f, "{}", self.to_display_string())
    }
}
impl Debug for Pattern {
    fn fmt(&self, f: &mut Formatter) -> Result<(), std::fmt::Error> {
        write!(f, "{}", self.to_debug_string())
    }
}
