use crate::{Attributes, Condition, Mutation};
use markdown_parser as md;

/// A clause contains prose, which is text to print to the screen, and conditions
/// that need to be fulfilled before the text is allowed to be shown
pub struct Clause {
    pub prose: String,
    pub conditions: Vec<Condition>,
    pub mutations: Vec<Mutation>,
}

impl Clause {
    /// Parse the content of a markdown quote block into a clause
    pub fn parse(lines: &[md::Line]) -> Self {
        let mut prose = String::new();
        let mut conditions = Vec::new();
        let mut mutations = Vec::new();
        for line in lines {
            for text in line {
                match text {
                    md::Text::Code(content) => {
                        let mut chars = content.chars();
                        let first_char = match chars.next() {
                            Some(v) => v,
                            None => panic!("Empty code block"),
                        };
                        let content: String = chars.collect();
                        match first_char {
                            '?' => conditions.push(Condition::parse(&content)),
                            '!' => mutations.push(Mutation::parse(&content)),
                            other => panic!("Unexpected starting char '{}'", other),
                        }
                    }
                    md::Text::Normal(string) => {
                        prose.push_str(string.trim());
                        prose.push(' ');
                    }
                    other => panic!("Unsupported text type in clause: {:?}", other),
                }
            }
        }
        Self {
            prose,
            conditions,
            mutations,
        }
    }

    pub fn check_conditions(&self, attributes: &mut Attributes) -> bool {
        for condition in &self.conditions {
            if !condition.evaluate(attributes) {
                return false;
            }
        }
        return true;
    }

    pub fn apply_mutations(&self, attributes: &mut Attributes) {
        for mutation in &self.mutations {
            mutation.apply(attributes);
        }
    }
}
