use crate::*;
use action_parser::Input;
use log::{debug, error};
use vagabond as vg;
use wetstring_http::*;

pub struct Game {
    locations: Vec<Location>,
    saves_directory: vg::Entry,
}

impl Game {
    pub fn new(game_directory: vg::Entry, saves_directory: vg::Entry) -> Self {
        let mut locations = Vec::new();
        for entry in vg::traverse_directory(game_directory).unwrap() {
            if entry.extension == "location" {
                locations.push(Location::parse(entry).unwrap());
            }
        }

        // Verify that all location mutations lead to a valid location
        let location_ids: Vec<&String> = locations.iter().map(|l| &l.identifier).collect();
        for location in &locations {
            for action in &location.actions {
                for clause in &action.clauses {
                    for mutation in &clause.mutations {
                        if mutation.attribute == "location" {
                            if !location_ids.contains(&&mutation.new_value) {
                                panic!(
                                    "{} contains dangling location pointer: {}",
                                    location.identifier, mutation.new_value
                                );
                            }
                        }
                    }
                }
            }
        }

        Game {
            locations,
            saves_directory,
        }
    }

    pub fn get_location(&self, location: &str) -> &Location {
        match self.locations.iter().find(|l| &l.identifier == location) {
            Some(v) => v,
            None => unreachable!("Player is in unknown location: {}", location),
        }
    }
}

impl RequestProcessor for Game {
    type Req = HttpRequest;
    type Res = HttpResponse;

    fn process_request(&self, request: &Self::Req) -> HttpResponse {
        let username = "alex";
        let mut save_file = match SaveFile::load(&self.saves_directory, username) {
            Ok(v) => v,
            Err(_) => {
                error!("Save file not found for user '{}'", username);
                return HttpResponse::client_error();
            }
        };

        // Set dynamic attributes
        save_file.attributes.set("$time_of_day", "day");

        let input = match request.target.query_parameters.get("action") {
            Some(v) => Input::parse(&v.replace('+', " ")),
            None => Input::parse(""),
        };
        let location_name = save_file.attributes.get_with_default("location", "start");
        let mut location = self.get_location(location_name);

        let mut action_prose = Vec::new();

        // Apply actions and mutations
        if let Some(action) = location.actions.iter().find(|a| a.matches(&input)) {
            debug!("Matched action: {}", action.name);
            for clause in &action.clauses {
                if clause.check_conditions(&mut save_file.attributes) {
                    clause.apply_mutations(&mut save_file.attributes);
                    action_prose.push(clause.prose.clone());
                }
            }
        } else {
            debug!("No match: {}", input);
        }

        let mut body_text = String::new();
        body_text.push_str(HTML_HEAD);

        if !action_prose.is_empty() {
            body_text.push_str(r#"<div style="height: 30px"></div>"#);
            for line in &action_prose {
                body_text.push_str(&format!("<p>{}</p>", line));
            }
            body_text.push_str(r#"<div style="height: 5px"></div>"#);

            let location_name = save_file.attributes.get("location");
            location = self.get_location(location_name);
        };

        // Print description for current location
        body_text.push_str(&format!("<h1>{}</h1>", location.name));
        for clause in &location.description {
            if clause.check_conditions(&mut save_file.attributes) {
                body_text.push_str(&format!("<p>{}</p>", clause.prose));
            }
        }

        body_text.push_str(HTML_TAIL);

        save_file.save();
        HttpResponse::new(Status::Success).with_utf8_body(&body_text)
    }
}

const HTML_HEAD: &str = r#"
<!DOCTYPE html>
<html>
<head>
  <link rel="preconnect" href="https://fonts.googleapis.com">
  <link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
  <link href="https://fonts.googleapis.com/css2?family=Fira+Code:wght@600&family=Vollkorn:wght@600&family=PT+Serif&family=Merriweather&display=swap" rel="stylesheet">
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <title>An Ode to the Ever-blowing Wind</title>
  <style>
    * {
      color: #022;
    }
    body {
        background: #ed2;
    }
    h1.title {
        color: #552;
        font-size: 36pt;
        font-family: 'PT Serif', serif;
        text-orientation: sideways;
        writing-mode: vertical-rl;
        white-space: nowrap;
        transform: rotate(180deg) translateX(100%);
        transform-origin: center;
        position: absolute;
        border-left: 2px solid #444;
        padding-left: 5px;
        margin: 0;
    }
    p {
        font-family: Merriweather, serif;
        font-size: 11pt;
    }
    div.container {
      max-width: 600px;
      margin: 0 auto;
      padding-left: 60px;
      height: 90%;
    }
    div.content {
        padding-left: 20px;
        padding-right: 20px;
        margin-left: 10px;
    }
    div.content > h1 {
        font-family: Vollkorn, serif;
        font-weight: 600;
    }
    input {
      font-family: 'Fira Code', monospace;
      font-weight: 600;
      font-size: 12pt;
      background-color: #ed6;
      border: none;
    }
    form.action {
      display: flex;
      width: 100%;
      border: 2px solid #022;
      border-radius: 4px;
    }
    input#action {
      padding: 2px 6px;
      margin: 0;
      width: 100%;
      text-transform: uppercase;
    }
    input#action:focus {
        outline: none;
    }
    input#submit {
      padding: 2px 6px;
      margin: 0;
      border-left: 2px solid #022;
    }
    input#submit:focus {
        outline: none;
        background-color: #eda;
    }

    /* Hide title on small screens */
    @media (max-width: 700px) {
      h1.title {
        display: none;
      }
      div.container {
        padding-left: 0px;
      }
      div.content {
        padding-left: 10px;
      }
      div.content > h1 {
        font-size: 20pt;
      }
    }
  </style>
</head>

<body>
  <div class="container">
    <h1 class="title">An Ode to the Ever-blowing Wind</h1>

    <div class="content">
"#;

const HTML_TAIL: &str = r#"
      <div style="height: 50px"></div>
      <form class="action" action="">
        <input id="action" name="action" type="text">
        <input id="submit" type="submit" value="->">
      </form>
    </div>
  </div>

  <div style="height: 10%"></div>
</body>
</html>
"#;
