use crate::{Action, Clause, Pattern};
use log::info;
use markdown_parser as md;
use vagabond as vg;

pub struct Location {
    pub name: String,
    pub identifier: String,
    pub description: Vec<Clause>,
    pub actions: Vec<Action>,
}

impl Location {
    pub fn parse(entry: vg::Entry) -> Result<Location, LocationParseError> {
        info!("Parsing {}", entry.name);

        let markdown = entry.read_as_utf8_string()?;
        let document =
            md::parse_heirarchical(&markdown).or(Err(LocationParseError::InvalidHeirarchy))?;
        if document.sections.len() != 1 {
            return Err(LocationParseError::InvalidHeirarchy);
        }
        let location = &document.sections[0];

        let name = md::line_to_string(&location.title);
        let identifier = entry.split_name().0;

        let description = {
            let mut clauses = Vec::new();
            for block in &location.content {
                if let md::Block::Quote(lines) = block {
                    clauses.push(Clause::parse(&lines));
                }
            }
            clauses
        };
        let actions = {
            let mut actions = Vec::new();
            for section in &location.sections {
                let name = md::line_to_string(&section.title);
                let mut clauses = Vec::new();
                let mut patterns = Vec::new();
                for block in &section.content {
                    match block {
                        md::Block::Quote(lines) => clauses.push(Clause::parse(&lines)),
                        md::Block::Paragraph(text) => {
                            if let Some(md::Text::Code(content)) = text.get(0) {
                                let mut chars = content.chars();
                                let first_char = match chars.next() {
                                    Some(v) => v,
                                    None => continue,
                                };
                                let content: String = chars.collect();
                                if first_char == '>' {
                                    patterns.push(Pattern::parse(&content).unwrap());
                                } else {
                                    panic!("Pattern doesn't start with a '>': {}", content)
                                }
                            }
                        }
                        _ => continue,
                    }
                }
                actions.push(Action {
                    name,
                    clauses,
                    patterns,
                })
            }
            actions
        };

        Ok(Location {
            name,
            identifier,
            description,
            actions,
        })
    }
}

#[derive(std::fmt::Debug)]
pub enum LocationParseError {
    InvalidPermissions,
    NotFound,
    InvalidHeirarchy,
}
impl From<vg::EntryReadError> for LocationParseError {
    fn from(error: vg::EntryReadError) -> Self {
        match error {
            vg::EntryReadError::NotFound => Self::NotFound,
            vg::EntryReadError::PermissionDenied => Self::InvalidPermissions,
        }
    }
}
