use std::collections::HashMap;
use std::path::{Path, PathBuf};
use vagabond as vg;

pub struct SaveFile {
    path: PathBuf,
    pub attributes: Attributes,
}

impl SaveFile {
    pub fn load<P: AsRef<Path>>(saves_directory: P, username: &str) -> Result<Self, ()> {
        let filename = format!("{}.save", username);
        let mut path = PathBuf::new();
        path.push(saves_directory);
        path.push(filename);
        let entry = match vg::Entry::from_path(&path) {
            Ok(v) => v,
            Err(vg::EntryReadError::NotFound) => return Err(()),
            Err(err) => panic!("{:?}", err),
        };
        let content = entry.read_as_utf8_string().unwrap();
        let attributes = Self::parse(&content)?;
        Ok(Self { path, attributes })
    }

    pub fn parse(content: &str) -> Result<Attributes, ()> {
        let mut attributes = Attributes::new();
        for line in content.lines() {
            if let Some((key, value)) = line.split_once('=') {
                attributes.set(key, value);
            }
        }
        Ok(attributes)
    }

    pub fn save(&self) {
        let mut content = String::new();
        for (key, value) in &self.attributes.map {
            content.push_str(key);
            content.push('=');
            content.push_str(value);
            content.push('\n');
        }
        std::fs::write(&self.path, &content).unwrap();
    }
}

pub struct Attributes {
    pub map: HashMap<String, String>,
}

impl Attributes {
    pub fn new() -> Self {
        Self {
            map: HashMap::new(),
        }
    }

    pub fn get(&mut self, name: &str) -> &String {
        self.get_with_default(name, "")
    }
    pub fn get_with_default(&mut self, name: &str, default: &str) -> &String {
        self.map
            .entry(name.to_string())
            .or_insert(default.to_string())
    }

    pub fn set(&mut self, name: &str, value: &str) {
        self.map.insert(name.to_string(), value.to_string());
    }
}
